import { z } from "zod";
import { prependForwardSlash } from "../core/path.js";
function createCollectionToGlobResultMap({
  globResult,
  contentDir
}) {
  const collectionToGlobResultMap = {};
  for (const key in globResult) {
    const keyRelativeToContentDir = key.replace(new RegExp(`^${contentDir}`), "");
    const segments = keyRelativeToContentDir.split("/");
    if (segments.length <= 1)
      continue;
    const collection = segments[0];
    const entryId = segments.slice(1).join("/");
    collectionToGlobResultMap[collection] ?? (collectionToGlobResultMap[collection] = {});
    collectionToGlobResultMap[collection][entryId] = globResult[key];
  }
  return collectionToGlobResultMap;
}
async function parseEntryData(collection, entry, collectionsConfig) {
  if (!("schema" in (collectionsConfig[collection] ?? {}))) {
    throw new Error(getErrorMsg.schemaDefMissing(collection));
  }
  const { schema } = collectionsConfig[collection];
  const parsed = z.object(schema).safeParse(entry.data, { errorMap });
  if (parsed.success) {
    return parsed.data;
  } else {
    const formattedError = new Error(
      [
        `Could not parse frontmatter in ${String(collection)} \u2192 ${String(entry.id)}`,
        ...parsed.error.errors.map((zodError) => zodError.message)
      ].join("\n")
    );
    formattedError.loc = {
      file: entry._internal.filePath,
      line: getFrontmatterErrorLine(
        entry._internal.rawData,
        String(parsed.error.errors[0].path[0])
      ),
      column: 1
    };
    throw formattedError;
  }
}
const flattenPath = (path) => path.join(".");
const errorMap = (error, ctx) => {
  if (error.code === "invalid_type") {
    const badKeyPath = JSON.stringify(flattenPath(error.path));
    if (error.received === "undefined") {
      return { message: `${badKeyPath} is required.` };
    } else {
      return { message: `${badKeyPath} should be ${error.expected}, not ${error.received}.` };
    }
  }
  return { message: ctx.defaultError };
};
function getFrontmatterErrorLine(rawFrontmatter, frontmatterKey) {
  const indexOfFrontmatterKey = rawFrontmatter.indexOf(`
${frontmatterKey}`);
  if (indexOfFrontmatterKey === -1)
    return 0;
  const frontmatterBeforeKey = rawFrontmatter.substring(0, indexOfFrontmatterKey + 1);
  const numNewlinesBeforeKey = frontmatterBeforeKey.split("\n").length;
  return numNewlinesBeforeKey;
}
const getErrorMsg = {
  schemaFileMissing: (collection) => `${collection} does not have a config. We suggest adding one for type safety!`,
  schemaDefMissing: (collection) => `${collection} needs a schema definition. Check your src/content/config!`
};
function createGetCollection({
  collectionToEntryMap,
  getCollectionsConfig
}) {
  return async function getCollection(collection, filter) {
    const lazyImports = Object.values(collectionToEntryMap[collection] ?? {});
    const collectionsConfig = await getCollectionsConfig();
    const entries = Promise.all(
      lazyImports.map(async (lazyImport) => {
        const entry = await lazyImport();
        const data = await parseEntryData(collection, entry, collectionsConfig);
        return {
          id: entry.id,
          slug: entry.slug,
          body: entry.body,
          collection: entry.collection,
          data
        };
      })
    );
    if (typeof filter === "function") {
      return (await entries).filter(filter);
    } else {
      return entries;
    }
  };
}
function createGetEntry({
  collectionToEntryMap,
  getCollectionsConfig
}) {
  return async function getEntry(collection, entryId) {
    var _a;
    const lazyImport = (_a = collectionToEntryMap[collection]) == null ? void 0 : _a[entryId];
    const collectionsConfig = await getCollectionsConfig();
    if (!lazyImport)
      throw new Error(`Ah! ${entryId}`);
    const entry = await lazyImport();
    const data = await parseEntryData(collection, entry, collectionsConfig);
    return {
      id: entry.id,
      slug: entry.slug,
      body: entry.body,
      collection: entry.collection,
      data
    };
  };
}
function createCollectionToPaths({
  getCollection
}) {
  return async function collectionToPaths(collection) {
    const entries = await getCollection(collection);
    return entries.map((entry) => ({
      params: { slug: entry.slug },
      props: entry
    }));
  };
}
function createRenderEntry({
  collectionToRenderEntryMap
}) {
  return async function renderEntry(entry) {
    var _a;
    const lazyImport = (_a = collectionToRenderEntryMap[entry.collection]) == null ? void 0 : _a[entry.id];
    if (!lazyImport)
      throw new Error(`${String(entry.collection)} \u2192 ${String(entry.id)} does not exist.`);
    const mod = await lazyImport();
    if (Array.isArray(mod == null ? void 0 : mod.collectedLinks) && "links" in (this ?? {})) {
      for (const link of mod.collectedLinks) {
        this.links.add({
          props: { rel: "stylesheet", href: prependForwardSlash(link) },
          children: ""
        });
      }
    }
    if (Array.isArray(mod == null ? void 0 : mod.collectedStyles) && "styles" in (this ?? {})) {
      for (const style of mod.collectedStyles) {
        this.styles.add({
          props: {},
          children: style
        });
      }
    }
    return {
      Content: mod.Content,
      headings: mod.getHeadings(),
      injectedFrontmatter: mod._internal.injectedFrontmatter
    };
  };
}
export {
  createCollectionToGlobResultMap,
  createCollectionToPaths,
  createGetCollection,
  createGetEntry,
  createRenderEntry,
  getErrorMsg,
  parseEntryData
};
