import { normalizePath } from "vite";
import glob from "fast-glob";
import * as fs from "node:fs/promises";
import * as fsSync from "node:fs";
import * as path from "node:path";
import { bold, cyan } from "kleur/colors";
import matter from "gray-matter";
import { info, warn } from "../core/logger/core.js";
import { appendForwardSlash, prependForwardSlash } from "../core/path.js";
import { contentFileExts, CONTENT_FLAG, VIRTUAL_MODULE_ID } from "./consts.js";
const CONTENT_BASE = "types.generated";
const CONTENT_FILE = CONTENT_BASE + ".mjs";
const CONTENT_TYPES_FILE = CONTENT_BASE + ".d.ts";
function astroContentPlugin({
  settings,
  logging
}) {
  const { root, srcDir } = settings.config;
  const paths = {
    cacheDir: new URL("./content/", srcDir),
    contentDir: new URL("./content/", srcDir),
    generatedInputDir: new URL("../../src/content/template/", import.meta.url),
    config: new URL("./content/config", srcDir)
  };
  let contentDirExists = false;
  let contentGenerator;
  const relContentDir = appendForwardSlash(
    prependForwardSlash(path.relative(settings.config.root.pathname, paths.contentDir.pathname))
  );
  const entryGlob = relContentDir + "**/*.{md,mdx}";
  const astroContentModContents = fsSync.readFileSync(new URL(CONTENT_FILE, paths.generatedInputDir), "utf-8").replace("@@CONTENT_DIR@@", relContentDir).replace("@@ENTRY_GLOB_PATH@@", entryGlob).replace("@@RENDER_ENTRY_GLOB_PATH@@", entryGlob).replace("@@COLLECTIONS_IMPORT_PATH@@", paths.config.pathname);
  const resolvedVirtualModuleId = "\0" + VIRTUAL_MODULE_ID;
  return [
    {
      name: "astro-content-virtual-module-plugin",
      resolveId(id) {
        if (id === VIRTUAL_MODULE_ID) {
          return resolvedVirtualModuleId;
        }
      },
      load(id) {
        if (id === resolvedVirtualModuleId) {
          return {
            code: astroContentModContents
          };
        }
      }
    },
    {
      name: "content-flag-plugin",
      enforce: "pre",
      async load(id) {
        const { pathname, searchParams } = new URL(id, "file://");
        if (searchParams.has(CONTENT_FLAG) && contentFileExts.some((ext) => pathname.endsWith(ext))) {
          const rawContents = await fs.readFile(pathname, "utf-8");
          const { content: body, data, matter: rawData } = parseFrontmatter(rawContents, pathname);
          const entryInfo = parseEntryInfo(pathname, { contentDir: paths.contentDir });
          if (entryInfo instanceof Error)
            return;
          return {
            code: `
export const id = ${JSON.stringify(entryInfo.id)};
export const collection = ${JSON.stringify(entryInfo.collection)};
export const slug = ${JSON.stringify(entryInfo.slug)};
export const body = ${JSON.stringify(body)};
export const data = ${JSON.stringify(data)};
export const _internal = {
	filePath: ${JSON.stringify(pathname)},
	rawData: ${JSON.stringify(rawData)},
};
`
          };
        }
      }
    },
    {
      name: "astro-fetch-content-plugin",
      async config() {
        try {
          await fs.stat(paths.contentDir);
          contentDirExists = true;
        } catch {
          return;
        }
        info(logging, "content", "Generating entries...");
        contentGenerator = await toGenerateContent({ logging, paths });
        await contentGenerator.init();
      },
      async configureServer(viteServer) {
        if (contentDirExists) {
          info(
            logging,
            "content",
            `Watching ${cyan(paths.contentDir.href.replace(root.href, ""))} for changes`
          );
          attachListeners();
        } else {
          viteServer.watcher.on("addDir", (dir) => {
            if (dir === paths.contentDir.pathname) {
              info(logging, "content", `Content dir found. Watching for changes`);
              contentDirExists = true;
              attachListeners();
            }
          });
        }
        function attachListeners() {
          viteServer.watcher.on(
            "add",
            (entry) => contentGenerator.queueEvent({ name: "add", entry })
          );
          viteServer.watcher.on(
            "addDir",
            (entry) => contentGenerator.queueEvent({ name: "addDir", entry })
          );
          viteServer.watcher.on(
            "change",
            (entry) => contentGenerator.queueEvent({ name: "change", entry })
          );
          viteServer.watcher.on(
            "unlink",
            (entry) => contentGenerator.queueEvent({ name: "unlink", entry })
          );
          viteServer.watcher.on(
            "unlinkDir",
            (entry) => contentGenerator.queueEvent({ name: "unlinkDir", entry })
          );
        }
      }
    }
  ];
}
const msg = {
  collectionAdded: (collection) => `${cyan(collection)} collection added`,
  entryAdded: (entry, collection) => `${cyan(entry)} added to ${bold(collection)}.`
};
async function toGenerateContent({
  logging,
  paths
}) {
  const contentTypes = {};
  let events = [];
  let debounceTimeout;
  let eventsSettled;
  const contentTypesBase = await fs.readFile(
    new URL(CONTENT_TYPES_FILE, paths.generatedInputDir),
    "utf-8"
  );
  async function init() {
    const pattern = new URL("./**/", paths.contentDir).pathname + "*.{md,mdx}";
    const entries = await glob(pattern);
    for (const entry of entries) {
      queueEvent({ name: "add", entry }, { shouldLog: false });
    }
    await eventsSettled;
  }
  async function onEvent(event, opts) {
    const shouldLog = (opts == null ? void 0 : opts.shouldLog) ?? true;
    if (event.name === "addDir" || event.name === "unlinkDir") {
      const collection = path.relative(paths.contentDir.pathname, event.entry);
      const isCollectionEvent = collection.split(path.sep).length === 1;
      if (!isCollectionEvent)
        return;
      switch (event.name) {
        case "addDir":
          addCollection(contentTypes, JSON.stringify(collection));
          if (shouldLog) {
            info(logging, "content", msg.collectionAdded(collection));
          }
          break;
        case "unlinkDir":
          removeCollection(contentTypes, JSON.stringify(collection));
          break;
      }
    } else {
      const fileType = getEntryType(event.entry, paths);
      if (fileType === "unknown") {
        warn(
          logging,
          "content",
          `${cyan(
            path.relative(paths.contentDir.pathname, event.entry)
          )} is not a supported file type. Skipping.`
        );
        return;
      }
      const entryInfo = parseEntryInfo(event.entry, paths);
      if (entryInfo instanceof Error)
        return;
      const { id, slug, collection } = entryInfo;
      const collectionKey = JSON.stringify(collection);
      const entryKey = JSON.stringify(id);
      switch (event.name) {
        case "add":
          if (!(collectionKey in contentTypes)) {
            addCollection(contentTypes, collectionKey);
          }
          if (!(entryKey in contentTypes[collectionKey])) {
            addEntry(contentTypes, collectionKey, entryKey, slug);
          }
          if (shouldLog) {
            info(logging, "content", msg.entryAdded(entryInfo.slug, entryInfo.collection));
          }
          break;
        case "unlink":
          if (collectionKey in contentTypes && entryKey in contentTypes[collectionKey]) {
            removeEntry(contentTypes, collectionKey, entryKey);
          }
          break;
        case "change":
          break;
      }
    }
  }
  function queueEvent(event, eventOpts) {
    if (!event.entry.startsWith(paths.contentDir.pathname))
      return;
    if (event.entry.endsWith(CONTENT_TYPES_FILE))
      return;
    events.push(onEvent(event, eventOpts));
    runEventsDebounced();
  }
  function runEventsDebounced() {
    eventsSettled = new Promise((resolve, reject) => {
      try {
        debounceTimeout && clearTimeout(debounceTimeout);
        debounceTimeout = setTimeout(async () => {
          await Promise.all(events);
          await writeContentFiles({
            contentTypes,
            paths,
            contentTypesBase
          });
          resolve();
        }, 50);
      } catch (e) {
        reject(e);
      }
    });
  }
  return { init, queueEvent };
}
function addCollection(contentMap, collectionKey) {
  contentMap[collectionKey] = {};
}
function removeCollection(contentMap, collectionKey) {
  delete contentMap[collectionKey];
}
function addEntry(contentTypes, collectionKey, entryKey, slug) {
  contentTypes[collectionKey][entryKey] = `{
  id: ${entryKey},
  slug: ${JSON.stringify(
    slug
  )},
  body: string,
  collection: ${collectionKey},
  data: InferEntrySchema<${collectionKey}>
}`;
}
function removeEntry(contentTypes, collectionKey, entryKey) {
  delete contentTypes[collectionKey][entryKey];
}
function parseEntryInfo(entryPath, { contentDir }) {
  const relativeEntryPath = normalizePath(path.relative(contentDir.pathname, entryPath));
  const collection = path.dirname(relativeEntryPath).split(path.sep).shift();
  if (!collection)
    return new Error();
  const id = path.relative(collection, relativeEntryPath);
  const slug = id.replace(path.extname(id), "");
  return {
    id,
    slug,
    collection
  };
}
function getEntryType(entryPath, paths) {
  const { dir, ext, name } = path.parse(entryPath);
  if ([".md", ".mdx"].includes(ext)) {
    return "content";
  } else if (path.join(dir, name) === paths.config.pathname) {
    return "config";
  } else {
    return "unknown";
  }
}
function parseFrontmatter(fileContents, filePath) {
  try {
    return matter(fileContents);
  } catch (e) {
    if (e.name === "YAMLException") {
      const err = e;
      err.id = filePath;
      err.loc = { file: e.id, line: e.mark.line + 1, column: e.mark.column };
      err.message = e.reason;
      throw err;
    } else {
      throw e;
    }
  }
}
function stringifyObjKeyValue(key, value) {
  return `${key}: ${value},
`;
}
async function writeContentFiles({
  paths,
  contentTypes,
  contentTypesBase
}) {
  let contentTypesStr = "";
  for (const collectionKey in contentTypes) {
    contentTypesStr += `${collectionKey}: {
`;
    for (const entryKey in contentTypes[collectionKey]) {
      const entry = contentTypes[collectionKey][entryKey];
      contentTypesStr += stringifyObjKeyValue(entryKey, entry);
    }
    contentTypesStr += `},
`;
  }
  contentTypesBase = contentTypesBase.replace("// @@ENTRY_MAP@@", contentTypesStr);
  contentTypesBase = contentTypesBase.replace("@@COLLECTIONS_IMPORT_PATH@@", paths.config.pathname);
  try {
    await fs.stat(paths.cacheDir);
  } catch {
    await fs.mkdir(paths.cacheDir);
  }
  await fs.writeFile(new URL(CONTENT_TYPES_FILE, paths.cacheDir), contentTypesBase);
}
export {
  astroContentPlugin
};
