import MagicString from "magic-string";
import parseImports from "parse-imports";
import { moduleIsTopLevelPage, walkParentInfos } from "../core/build/graph.js";
import { getPageDataByViteID } from "../core/build/internal.js";
import { normalizeFilename } from "../vite-plugin-utils/index.js";
import { getStylesForURL } from "../core/render/dev/css.js";
import { pathToFileURL } from "url";
import { createViteLoader } from "../core/module-loader/vite.js";
import {
  contentFileExts,
  DELAYED_ASSET_FLAG,
  LINKS_PLACEHOLDER,
  STYLES_PLACEHOLDER,
  VIRTUAL_MODULE_ID
} from "./consts.js";
function isDelayedAsset(url) {
  return url.searchParams.has(DELAYED_ASSET_FLAG) && contentFileExts.some((ext) => url.pathname.endsWith(ext));
}
function astroDelayedAssetPlugin({
  settings,
  mode
}) {
  let devModuleLoader;
  return {
    name: "astro-delayed-asset-plugin",
    enforce: "pre",
    configureServer(server) {
      if (mode === "dev") {
        devModuleLoader = createViteLoader(server);
      }
    },
    load(id) {
      const url = new URL(id, "file://");
      if (isDelayedAsset(url)) {
        const code = `
					export { Content, getHeadings, _internal } from ${JSON.stringify(url.pathname)};
					export const collectedLinks = ${JSON.stringify(LINKS_PLACEHOLDER)};
					export const collectedStyles = ${JSON.stringify(STYLES_PLACEHOLDER)};
				`;
        return { code };
      }
    },
    async transform(code, id, options) {
      var _a;
      if (!(options == null ? void 0 : options.ssr))
        return;
      const url = new URL(id, "file://");
      if (devModuleLoader && isDelayedAsset(url)) {
        const { pathname } = url;
        if (!((_a = devModuleLoader.getModuleById(pathname)) == null ? void 0 : _a.ssrModule)) {
          await devModuleLoader.import(pathname);
        }
        const { stylesMap, urls } = await getStylesForURL(
          pathToFileURL(pathname),
          devModuleLoader,
          "development"
        );
        return {
          code: code.replace(JSON.stringify(LINKS_PLACEHOLDER), JSON.stringify([...urls])).replace(JSON.stringify(STYLES_PLACEHOLDER), JSON.stringify([...stylesMap.values()]))
        };
      }
      if (id.endsWith(".astro")) {
        let renderEntryImportName = getRenderEntryImportName(
          await parseImports(escapeViteEnvReferences(code))
        );
        if (!renderEntryImportName)
          return;
        const s = new MagicString(code, {
          filename: normalizeFilename({ fileName: id, projectRoot: settings.config.root })
        });
        s.prepend(
          `import { renderEntry as $$renderEntry } from ${JSON.stringify(VIRTUAL_MODULE_ID)};
`
        );
        const frontmatterPreamble = "$$createComponent(async ($$result, $$props, $$slots) => {";
        const indexOfFrontmatterPreamble = code.indexOf(frontmatterPreamble);
        if (indexOfFrontmatterPreamble < 0)
          return;
        s.appendLeft(
          indexOfFrontmatterPreamble + frontmatterPreamble.length,
          `
let ${renderEntryImportName} = $$renderEntry.bind($$result);
`
        );
        return {
          code: s.toString(),
          map: s.generateMap()
        };
      }
    }
  };
}
function astroBundleDelayedAssetPlugin({
  internals
}) {
  return {
    name: "astro-bundle-delayed-asset-plugin",
    async generateBundle(_options, bundle) {
      var _a;
      for (const [_, chunk] of Object.entries(bundle)) {
        if (chunk.type === "chunk" && chunk.code.includes(LINKS_PLACEHOLDER)) {
          for (const id of Object.keys(chunk.modules)) {
            for (const [pageInfo, depth, order] of walkParentInfos(id, this)) {
              if (moduleIsTopLevelPage(pageInfo)) {
                const pageViteID = pageInfo.id;
                const pageData = getPageDataByViteID(internals, pageViteID);
                if (!pageData)
                  continue;
                const entryDeferredCss = (_a = pageData.contentDeferredCss) == null ? void 0 : _a.get(id);
                if (!entryDeferredCss)
                  continue;
                chunk.code = chunk.code.replace(
                  JSON.stringify(LINKS_PLACEHOLDER),
                  JSON.stringify([...entryDeferredCss])
                );
              }
            }
          }
        }
      }
    }
  };
}
function getRenderEntryImportName(parseImportRes) {
  var _a;
  for (const imp of parseImportRes) {
    if (imp.moduleSpecifier.value === VIRTUAL_MODULE_ID && ((_a = imp.importClause) == null ? void 0 : _a.named)) {
      for (const namedImp of imp.importClause.named) {
        if (namedImp.specifier === "renderEntry") {
          return namedImp.binding;
        }
      }
    }
  }
  return void 0;
}
function escapeViteEnvReferences(code) {
  return code.replace(/import\.meta/g, "import\\u002Emeta");
}
export {
  astroBundleDelayedAssetPlugin,
  astroDelayedAssetPlugin
};
